; FaceBoard					(c) 2015 Cygnus2
; www.cygnus2.net
;
;	**********************************************************************************************************
;	**													**
;	**				S O U R C E    C O D E    R E V I S I O N				**
;	**													**
;	**********************************************************************************************************
;	**													**
;	** 					Created by:  Vladimir RYKL					**
;	**													**
;	**********************************************************************************************************
;	**	   *		*										**
;	** Version *    Date 	*				Description					**
;	** 	   *		*										**
;	**********************************************************************************************************
;	**  V0.01  * 18/02/2015	* First release	- MCU PIC16F630	used instead of PIC16F505			**
;	** 	   *		*										**
;	** 	   *		* ROM occupation: 45.4% (465 of 1024 words used)				**
;	**********************************************************************************************************
;	**  V0.02  * 14/03/2015	* Game start by button hold							**
;	** 	   *		*										**
;	** 	   *		* ROM occupation: 44.8% (459 of 1024 words used)				**
;	**********************************************************************************************************
;	**  V0.03  * 16/03/2015	* All effects controlled as automatically as by sound				**
;	** 	   *		*										**
;	** 	   *		* ROM occupation: 46.7% (478 of 1024 words used)				**
;	**********************************************************************************************************
;	**  V1.00  * 09/06/2015	* Adapted to the use of MCU PIC16F610, otherwise the same as V0.03		**
;	** 	   *		*										**
;	** 	   *		* ROM occupation: 46.6% (477 of 1024 words used)				**
;	**********************************************************************************************************
;
	list      p=16F610		; list directive to define processor
	#include <p16F610.inc>		; processor specific variable definitions
;
;************************************************************************************************************************
;		R E G I S T E R     A N D     B I T     D E F I N I T I O N
;************************************************************************************************************************
; Program control register
reg_ProgCtrl		EQU	07FH				; Program control register 1
#define		bit_GameActive		reg_ProgCtrl,003H	; Game flag
#define		bit_TimeWindow		reg_ProgCtrl,002H	; Time window
#define		bit_Hold		reg_ProgCtrl,001H	; Hold flag (0 - key released, 1 - key hold)
#define		bit_Key			reg_ProgCtrl,000H	; Key flag (0 - key released, 1 - key pressed)
; Delay control register
reg_DelayCtrl		EQU	07EH				; Delay expiration flags
#define		bit_LongDelay3		reg_DelayCtrl,007H	; Long delay 1 expiration flag
#define		bit_LongDelay2		reg_DelayCtrl,006H	; Long delay 2 expiration flag
#define		bit_LongDelay1		reg_DelayCtrl,005H	; Long delay 3 expiration flag
#define		bit_ShortDelay4		reg_DelayCtrl,004H	; Short delay 1 expiration flag
#define		bit_ShortDelay3		reg_DelayCtrl,003H	; Short delay 2 expiration flag
#define		bit_ShortDelay2		reg_DelayCtrl,002H	; Short delay 3 expiration flag
#define		bit_ShortDelay1		reg_DelayCtrl,001H	; Short delay 4 expiration flag
#define		bit_BasicDelay		reg_DelayCtrl,000H	; Basic delay expiurstion flag
; Delay counter registers
reg_LongDelay3H		EQU	07DH				; Long delay counter register (high & low byte)
reg_LongDelay3L		EQU	07CH				; //
reg_LongDelay2H		EQU	07BH				; //
reg_LongDelay2L		EQU	07AH				; //
reg_LongDelay1H		EQU	079H				; //
reg_LongDelay1L		EQU	078H				; //
reg_ShortDelay4		EQU	077H				; Short delay counter register
reg_ShortDelay3		EQU	076H				; //
reg_ShortDelay2		EQU	075H				; //
reg_ShortDelay1		EQU	074H				; //
; Key (switch) registers
reg_KeyCode		EQU	073H				; Key code register
reg_KeyCdStk		EQU	072H				; Key code stack
reg_KeySamp		EQU	071H				; Key sampling register
; Function phase registers
reg_F03Phase		EQU	070H				; Function phase registers
reg_F04Phase		EQU	06FH				; //
reg_F05Phase		EQU	06EH				; //
; Temporary registers
reg_Temp1		EQU	06DH				; //
reg_Temp2		EQU	06CH				; //
reg_Temp3		EQU	06BH				; //
reg_Temp4		EQU	06AH				; //
; LED driver
reg_DriverRow2		EQU	069H				; 2 x 8 LED multiplex
reg_DriverRow1		EQU	068H				; //
reg_FaceCode		EQU	067H
reg_FacePeriod		EQU	066H
reg_SeqCode		EQU	065H
reg_SeqSymbPtr		EQU	064H
; Pseudo-random value generator
reg_RND1		EQU	063H				; Pseudo-random accumulator
reg_RND2		EQU	062H				; //
reg_RND3		EQU	061H				; //
reg_RND4		EQU	060H				; //
reg_Random		EQU	05FH				; Result register
;
reg_GameSpeed		EQU	05EH				; Game speed
reg_OKCounter		EQU	05DH				; OK counter
;************************************************************************************************************************
;		P O R T     D E F I N I T I O N
;************************************************************************************************************************
;
#define		port_ROW1		PORTA,004H		; Row 1 driver
#define		port_ROW2		PORTA,005H		; Row 2 driver
#define		port_COL1		PORTC,001H		; Column 1 driver
#define		port_COL2		PORTC,003H		; Column 2 driver
#define		port_COL3		PORTC,005H		; Column 3 driver
#define		port_COL4		PORTC,002H		; Column 4 driver
#define		port_COL5		PORTC,004H		; Column 5 driver
#define		port_COL6		PORTC,000H		; Column 6 driver
#define		port_COL7		PORTA,001H		; Column 7 driver
#define		port_COL8		PORTA,000H		; Column 8 driver
#define		port_KEY		PORTA,003H		; Key (switch) input
#define		port_MIC		PORTA,002H		; Microphone input
;
;************************************************************************************************************************
;		C O N S T A N T     D E F I N I T I O N
;************************************************************************************************************************
;
con_BasicSpeed		EQU	003H				; Initial game speed
con_SpeedStep		EQU	001H				; Constant for game speed decreasing
con_SpeedLimit		EQU	001H				; Highest game speed
con_OKValue		EQU	003H				; OK counter value for game speed change
;
;************************************************************************************************************************
;		P I C 1 6 F 6 3 0     C O N F I G U R A T I O N     F U S E S
;************************************************************************************************************************
;
;	__CONFIG   _CP & _CPD & _BODEN_OFF & _MCLRE_OFF & _WDT_OFF & _PWRTE_ON & _INTRC_OSC_NOCLKOUT
	__CONFIG   _BOD_OFF & _IOSCFS_4MHZ & _CP_OFF & _MCLRE_OFF & _PWRTE_ON & _WDT_ON  & _INTRC_OSC_NOCLKOUT
;
; MCLR configured as I/O
;
;************************************************************************************************************************
;*		A P P L I C A T I O N     B E G I N									*
;************************************************************************************************************************
;
		ORG	0000H				; RESET vector
;
sa_RESET:	GOTO	sa_INIT				; Skip interrupt service routine
;
;************************************************************************************************************************
;*		I N T E R R U P T     S U B R O U T I N E S								*
;************************************************************************************************************************
;
		ORG	0004H				; Interrupt vector
; TMR0 interrupt service routine - TMR0 interrupt occurs every 1024 usec
		BCF	INTCON,T0IF			; Clear TMR0 interrupt flag
; Basic delay
		BSF	bit_BasicDelay			; Set basic delay expiration flag
; Short delay 1 (watchdog reset)
		DECFSZ	reg_ShortDelay1,F		; Delay expired ?
		GOTO	sa_TMR0A			; // NO !
		BSF	bit_ShortDelay1			; // YES ! - set expiration flag
; Short delay 2 (key samplig period)
sa_TMR0A:	DECFSZ	reg_ShortDelay2,F		; Delay expired ?
		GOTO	sa_TMR0B			; // NO !
		BSF	bit_ShortDelay2			; // YES ! - set expiration flag
; Short delay 3
sa_TMR0B:	DECFSZ	reg_ShortDelay3,F		; Delay expired ?
		GOTO	sa_TMR0C			; // NO !
		BSF	bit_ShortDelay3			; // YES ! - set expiration flag
; Short delay 4
sa_TMR0C:	DECFSZ	reg_ShortDelay4,F		; Delay expired ?
		GOTO	sa_TMR0D			; // NO !
		BSF	bit_ShortDelay4			; // YES ! - set expiration flag
; Long delay 1 (face duration)
sa_TMR0D:	DECFSZ	reg_LongDelay1L,F		; Delay expired ?
		GOTO	sa_TMR0E			; // NO !
		DECFSZ	reg_LongDelay1H,F		; //
		GOTO	sa_TMR0E			; // NO !
		BSF	bit_LongDelay1			; // YES ! - set expiration flag
; Long delay 2 (key hold)
sa_TMR0E:	DECFSZ	reg_LongDelay2L,F		; Delay expired ?
		GOTO	sa_TMR0F			; // NO !
		DECFSZ	reg_LongDelay2H,F		; //
		GOTO	sa_TMR0F			; // NO !
		BSF	bit_LongDelay2			; // YES ! - set expiration flag
; Long delay 3
sa_TMR0F:	DECFSZ	reg_LongDelay3L,F		; Delay expired ?
		GOTO	sa_TMR0G			; // NO !
		DECFSZ	reg_LongDelay3H,F		; //
		GOTO	sa_TMR0G			; // NO !
		BSF	bit_LongDelay3			; // YES ! - set expiration flag
sa_TMR0G:	RETFIE					; Exit interrupt service routine
;
;****************************************************************************************************************
;*		S Y S T E M     I N I T I A L I Z A T I O N							*
;****************************************************************************************************************
;
;		---------------------
;		HARDWARE MODULE SETUP
;		---------------------
;
; I/O port setup
sa_INIT:	BCF	STATUS,RP0			; --> REGISTER BANK 0
		CLRF	PORTA				; Clear port driver
		CLRF	PORTC				; //
		BSF	STATUS,RP0			; --> REGISTER BANK 1
		MOVLW	00CH				; RA2, RA3 = input, RA0, RA1, RA4, RA5 = output
		MOVWF	TRISA				; //
		MOVLW	000H				; RC0 - RC5 = output
		MOVWF	TRISC				; //
		CLRF	ANSEL				; //
; TMR0 setup, clock source Fosc/4 ~ 1 MHz)
		MOVLW	0D1H				; TMR0 prescaler 1:4
		MOVWF	OPTION_REG			; //
		BCF	STATUS,RP0			; --> REGISTER BANK 0
		CLRF	TMR0				; Clear TMR0
;
;		---------------------------
;		MCU REGISTER INITIALIZATION
;		---------------------------
;
		CLRF	reg_ProgCtrl			; Clear program control register
		MOVLW	0FFH				; Set all delay expiration flags
		MOVWF	reg_DelayCtrl			; //
;
		CLRF	reg_KeyCode			; Clear key code register
		CLRF	reg_KeyCdStk			; Clear key code stack
		MOVLW	055H				; Initialize key sampling register
		MOVWF	reg_KeySamp			; //
;
		CLRF	reg_F03Phase			; Clear all function phase registers
		CLRF	reg_F04Phase			; //
		CLRF	reg_F05Phase			; //
;
		CLRF	reg_SeqCode			; First sequence
		CLRF	reg_DriverRow1			; Initialize LED driver (all LEDs turned off)
		CLRF	reg_DriverRow2			; //
;
		CLRF	reg_RND1			; Clear pseudo-random generator
		CLRF	reg_RND2			; //
		CLRF	reg_RND3			; //
		CLRF	reg_RND4			; //
;
		BSF	INTCON,T0IE			; TMR0 interrupt enable
		BSF	INTCON,GIE			; Global interrupt enable
;
;************************************************************************************************************************
;*		M A I N     P R O G R A M     L O O P									*
;************************************************************************************************************************
;
sa_MAINLOOP
;
;////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
; FUNCTION 01: Watchdog reset every 7 msec (watchdog time time-out period Twdt = 10 - 30 msec)
		BTFSS	bit_ShortDelay1			; Watchdog reset delay expired ?
		GOTO	sa_F01EXIT			; // NO ! - exit function
		BCF	bit_ShortDelay1			; // YES ! - clear expiration flag
		MOVLW	005H				; // Reinitialize delay for 5 msec
		MOVWF	reg_ShortDelay1			; //
		CLRWDT					; Reset watchdog timer
;------------------------------------------------------------------------------------------------------------------------
sa_F01EXIT						; FUNCTION 01 delimiter
;////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
;
;////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
; FUNCTION 02: Key (switch) sampling - 8 subsequent samples have to be equal to recognize valid key code
; Code 0x00 - switch released / code 0x01 - switch pressed
; Key sampling period about 5 msec, key response 8 x 5 = 40 msec
		BTFSS	bit_ShortDelay2			; Key sampling period expired ?
		GOTO	sa_F02EXIT			; // NO ! - exit function
		BCF	bit_ShortDelay2			; // YES ! - clear expiration flag
		MOVLW	005H				; Reinitialize delay for 5 msec
		MOVWF	reg_ShortDelay2			; //
;
		MOVLW	000H				; Store current key status into temporary register
		BTFSS	port_KEY			; //
		MOVLW	001H				; //
		MOVWF	reg_Temp1			; //
;
		XORWF	reg_KeyCdStk,W			; Compare two consecutive key code samples
		BCF	STATUS,C			; If two consecutive samples equal, shift in '0', otherwise	
		BTFSS	STATUS,Z			; // shift in '1'
		BSF	STATUS,C			; //
		RLF	reg_KeySamp,F			; //
;
		MOVF	reg_Temp1,W			; Restore current key status from temporary register
		MOVF	reg_KeySamp,F			; Last 8 samples equal ?
		BTFSC	STATUS,Z			; //
		MOVWF	reg_KeyCode			; // YES ! - update key code with new value
		MOVWF	reg_KeyCdStk			; //
;
		MOVLW	000H				; reg_KeyCode = 0x00 --> bit_Key = 0 & bit_Hold = 0
		XORWF	reg_KeyCode,W			; reg_KeyCode <> 0x00 --> bit_Key = 1
		BTFSS	STATUS,Z			; //
		GOTO	sa_F0201			; //
		BCF	bit_Key				; bit_Key = 0
		BCF	bit_Hold			; bit_Hold = 0
		BCF	bit_LongDelay2			; Clear expiration flag
		MOVLW	004H				; Initialize hold timer
		MOVWF	reg_LongDelay2H			; //
		MOVLW	0C8H				; //
		MOVWF	reg_LongDelay2L			; //
		GOTO	sa_F02EXIT			; //
sa_F0201:	BSF	bit_Key				; bit_Key = 1
		BTFSC	bit_LongDelay2			; Hold delay expired ?
		BSF	bit_Hold			; // YES ! - bit_Hold = 1
;------------------------------------------------------------------------------------------------------------------------
sa_F02EXIT						; FUNCTION 02 delimiter
;////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
;
;////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
; FUNCTION 03: LED driver control
		BTFSS	bit_BasicDelay			; Row switching delay expired ?
		GOTO	sa_F03EXIT			; // NO ! - exit function
		BCF	bit_BasicDelay			; // YES ! - clear expiration flag
;
		MOVLW	high(sa_F03PHASE)		; Function phase switch
		MOVWF	PCLATH				; //
		MOVF	reg_F03Phase,W			; //
		ADDWF	PCL,F				; //
sa_F03PHASE:	GOTO	sa_F03PHASE0			; //
sa_F03PHEND:	GOTO	sa_F03PHASE1			; //
;****** MPASM compiler will issue an error message because of PCLATH overrun !!! ****************************************
	IF	high(sa_F03PHASE) - high(sa_F03PHEND) != 0
		ERROR	"The 256B-boundary of program memory exceeding not allowed !!!"
	ENDIF
;************************************************************************************************************************
sa_F03PHASE0:	BCF	port_ROW2			; Turn off ROW2 driver before turning on ROW1 driver
		BTFSS	reg_DriverRow1,000H		; reg_DriverRow1 -> ports COL1 - COL8
		BSF	port_COL1			; Bit no. 0
		BTFSC	reg_DriverRow1,000H		; //
		BCF	port_COL1			; //
		BTFSS	reg_DriverRow1,001H		; Bit no. 1
		BSF	port_COL2			; //
		BTFSC	reg_DriverRow1,001H		; //
		BCF	port_COL2			; //
		BTFSS	reg_DriverRow1,002H		; Bit no. 2
		BSF	port_COL3			; //
		BTFSC	reg_DriverRow1,002H		; //
		BCF	port_COL3			; //
		BTFSS	reg_DriverRow1,003H		; Bit no. 3
		BSF	port_COL4			; //
		BTFSC	reg_DriverRow1,003H		; //
		BCF	port_COL4			; //
		BTFSS	reg_DriverRow1,004H		; Bit no. 4
		BSF	port_COL5			; //
		BTFSC	reg_DriverRow1,004H		; //
		BCF	port_COL5			; //
		BTFSS	reg_DriverRow1,005H		; Bit no. 5
		BSF	port_COL6			; //
		BTFSC	reg_DriverRow1,005H		; //
		BCF	port_COL6			; //
		BTFSS	reg_DriverRow1,006H		; Bit no. 6
		BSF	port_COL7			; //
		BTFSC	reg_DriverRow1,006H		; //
		BCF	port_COL7			; //
		BTFSS	reg_DriverRow1,007H		; Bit no. 7
		BSF	port_COL8			; //
		BTFSC	reg_DriverRow1,007H		; //
		BCF	port_COL8			; //
		BSF	port_ROW1			; Turn on ROW1 driver
		INCF	reg_F03Phase,F			; Phase 1
		GOTO	sa_F03EXIT			; Exit function
;------------------------------------------------------------------------------------------------------------------------
sa_F03PHASE1:	BCF	port_ROW1			; Turn off ROW1 driver before turning on ROW2 driver
		BTFSS	reg_DriverRow2,000H		; reg_DriverRow2 -> ports COL1 - COL8
		BSF	port_COL1			; Bit no. 0
		BTFSC	reg_DriverRow2,000H		; //
		BCF	port_COL1			; //
		BTFSS	reg_DriverRow2,001H		; Bit no. 1
		BSF	port_COL2			; //
		BTFSC	reg_DriverRow2,001H		; //
		BCF	port_COL2			; //
		BTFSS	reg_DriverRow2,002H		; Bit no. 2
		BSF	port_COL3			; //
		BTFSC	reg_DriverRow2,002H		; //
		BCF	port_COL3			; //
		BTFSS	reg_DriverRow2,003H		; Bit no. 3
		BSF	port_COL4			; //
		BTFSC	reg_DriverRow2,003H		; //
		BCF	port_COL4			; //
		BTFSS	reg_DriverRow2,004H		; Bit no. 4
		BSF	port_COL5			; //
		BTFSC	reg_DriverRow2,004H		; //
		BCF	port_COL5			; //
		BTFSS	reg_DriverRow2,005H		; Bit no. 5
		BSF	port_COL6			; //
		BTFSC	reg_DriverRow2,005H		; //
		BCF	port_COL6			; //
		BTFSS	reg_DriverRow2,006H		; Bit no. 6
		BSF	port_COL7			; //
		BTFSC	reg_DriverRow2,006H		; //
		BCF	port_COL7			; //
		BTFSS	reg_DriverRow2,007H		; Bit no. 7
		BSF	port_COL8			; //
		BTFSC	reg_DriverRow2,007H		; //
		BCF	port_COL8			; //
		BSF	port_ROW2			; Turn on ROW2 driver
		CLRF	reg_F03Phase			; Phase 0
		GOTO	sa_F03EXIT			; Exit function
;------------------------------------------------------------------------------------------------------------------------
sa_F03EXIT						; FUNCTION 03 delimiter
;////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
;
;////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
; FUNCTION 04: Face sequence generator
		MOVLW	high(sa_F04PHASE)		; Function phase switch
		MOVWF	PCLATH				; //
		MOVF	reg_F04Phase,W			; //
		ADDWF	PCL,F				; //
sa_F04PHASE:	GOTO	sa_F04PHASE0			; //
		GOTO	sa_F04PHASE1			; //
		GOTO	sa_F04PHASE2			; //
		GOTO	sa_F04PHASE3			; //
sa_F04PHEND:	GOTO	sa_F04PHASE4			; //
;****** MPASM compiler will issue an error message because of PCLATH overrun !!! ****************************************
	IF	high(sa_F04PHASE) - high(sa_F04PHEND) != 0
		ERROR	"The 256B-boundary of program memory exceeding not allowed !!!"
	ENDIF
;************************************************************************************************************************
sa_F04PHASE0:	CALL	sa_SEQDECODE			; Initialize sequence pointer
		MOVWF	reg_SeqSymbPtr			; //
		INCF	reg_F04Phase,F			; Phase 1
		GOTO	sa_F04EXIT			; Exit function
;------------------------------------------------------------------------------------------------------------------------
sa_F04PHASE1:	BTFSS	bit_GameActive			; Game active ?
		GOTO	sa_F04PH110			; // NO !
; Game active
		BCF	bit_TimeWindow			; // YES ! - deactivate time window
		MOVLW	050H				; Random value = 0x50H ?
		XORWF	reg_Random,W			; //
		BTFSS	STATUS,Z			; //
		GOTO	sa_F04PH101			; //
		MOVLW	00FH				; // YES ! - face code 0x0F, turn-off face conditionaly
		MOVWF	reg_FaceCode			; //
		MOVF	reg_GameSpeed,W			; Face period (variable)
		MOVWF	reg_FacePeriod			; //
		BSF	bit_TimeWindow			; Activate time window
		INCF	reg_F04Phase,F			; Phase 2
		GOTO	sa_F04PH190			; //
;
sa_F04PH101:	CALL	sa_SEQUENCE			; Get first / next face code
		XORLW	000H				; EOS found ?
		BTFSS	STATUS,Z			; //
		GOTO	sa_F04PH102			; // NO !
		CLRF	reg_F04Phase			; // YES ! - repeat sequence --> phase 0
		GOTO	sa_F04EXIT			; Exit function
;
sa_F04PH102:	INCF	reg_SeqSymbPtr,F		; Increment symbol pointer (next face)
		MOVWF	reg_FaceCode			; Decode face code (low nibble) and face period (high nibble)
		MOVLW	00FH				; //
		ANDWF	reg_FaceCode,F			; //
		MOVF	reg_GameSpeed,W			; Face period (variable)
		MOVWF	reg_FacePeriod			; //
		INCF	reg_F04Phase,F			; Phase 2
		GOTO	sa_F04PH190			; //
; Game inactive
sa_F04PH110:	CALL	sa_SEQUENCE			; Get first / next face code
		XORLW	000H				; EOS found ?
		BTFSS	STATUS,Z			; //
		GOTO	sa_F04PH111			; // NO !
		CLRF	reg_F04Phase			; // YES ! - repeat sequence --> phase 0
		GOTO	sa_F04EXIT			; Exit function
;
sa_F04PH111:	INCF	reg_SeqSymbPtr,F		; Increment symbol pointer (next face)
		MOVWF	reg_FaceCode			; Decode face code (low nibble) and face period (high nibble)
		SWAPF	reg_FaceCode,W			; //
		ANDLW	00FH				; //
		MOVWF	reg_FacePeriod			; //
		MOVLW	00FH				; //
		ANDWF	reg_FaceCode,F			; //
		INCF	reg_F04Phase,F			; Phase 2
; Update LED driver registers
sa_F04PH190:	RLF	reg_FaceCode,W			; reg_FaceCode * 2
		ANDLW	0FEH				; // ignore C status flag
		MOVWF	reg_Temp1			; //
		CALL	sa_FACEDECODE			; Decoding of reg_FaceCode --> reg_DriverRow1
		MOVWF	reg_DriverRow1			; //
		INCF	reg_Temp1,F			; reg_FaceCode * 2 + 1
		CALL	sa_FACEDECODE			; Decoding of reg_FaceCode --> reg_DriverRow2
		MOVWF	reg_DriverRow2			; //
;
		BCF	bit_LongDelay1			; Clear expiration flag
		MOVF	reg_FacePeriod,W		; Initialize face timer
		MOVWF	reg_LongDelay1H			; //
		MOVLW	0FFH				; //
		MOVWF	reg_LongDelay1L			; //
		GOTO	sa_F04EXIT			; Exit function
;------------------------------------------------------------------------------------------------------------------------
sa_F04PHASE2:	BTFSC	bit_GameActive			; Game active ?
		GOTO	sa_F04PH201			; // YES !
; Game inactive
		BTFSS	port_MIC			; Sound detected ?
		GOTO	sa_F04PH290			; YES !
		BTFSC	bit_LongDelay1			; Face period expired ?
		DECF	reg_F04Phase,F			; // YES ! - phase 1
		GOTO	sa_F04EXIT			; Exit function
; Game active
sa_F04PH201:	BTFSC	bit_TimeWindow			; Time window active ?
		GOTO	sa_F04PH202			; // YES !
; Time window inactive
		BTFSS	port_MIC			; Sound detected ?
		GOTO	sa_F04PH210			; // YES !
		BTFSC	bit_LongDelay1			; Face period expired ?
		DECF	reg_F04Phase,F			; // YES ! - phase 1
		GOTO	sa_F04EXIT			; Exit function
; Time window active
sa_F04PH202:	BTFSC	bit_LongDelay1			; Face period expired ?
		GOTO	sa_F04PH210			; // YES !
		BTFSS	port_MIC			; Sound detected ?
		GOTO	sa_F04PH220			; // YES !
		GOTO	sa_F04EXIT			; Exit function
;
sa_F04PH210:	MOVLW	con_BasicSpeed			; Reinitialize game speed
		MOVWF	reg_GameSpeed			; //
		CLRF	reg_OKCounter			; Clear OK counter
		MOVLW	095H				; NOK face
		MOVWF	reg_DriverRow1			; //
		MOVLW	0D5H				; //
		MOVWF	reg_DriverRow2			; //
		GOTO	sa_F04PH230			; //
;
sa_F04PH220:	INCF	reg_OKCounter,F			; Increment OK counter
		MOVLW	con_OKValue			; OK counter limit achieved (0x03) ?
		SUBWF	reg_OKCounter,W			; //
		BTFSS	STATUS,C			; //
		GOTO	sa_F04PH221			; // NO ! - no change of game speed
		CLRF	reg_OKCounter			; // YES ! - clear OK counter
		MOVLW	con_SpeedStep			; // Increase game speed
		SUBWF	reg_GameSpeed,F			; //
		MOVLW	con_SpeedLimit			; Game speed limit exceeded (0x02) ?
		SUBWF	reg_GameSpeed,W			; //
		MOVLW	con_SpeedLimit			; //
		BTFSS	STATUS,C			; //
		MOVWF	reg_GameSpeed			; // YES ! - set game speed limit (0x02)
sa_F04PH221:	MOVLW	06DH				; OK face
		MOVWF	reg_DriverRow1			; //
		MOVLW	02DH				; //
		MOVWF	reg_DriverRow2			; //
;
sa_F04PH230:	BCF	bit_LongDelay1			; Clear expiration flag
		MOVLW	005H				; Initialize face timer
		MOVWF	reg_LongDelay1H			; //
		MOVLW	0FFH				; //
		MOVWF	reg_LongDelay1L			; //
sa_F04PH240:	INCF	reg_F04Phase,F			; Phase 3
		GOTO	sa_F04EXIT			; Exit function
; Soun detected
sa_F04PH290:	BCF	bit_LongDelay1			; Clear expiration flag
		MOVF	reg_FacePeriod,W		; Initialize face timer
		MOVWF	reg_LongDelay1H			; //
		MOVLW	0FFH				; //
		MOVWF	reg_LongDelay1L			; //
		MOVLW	004H				; Phase 4
		MOVWF	reg_F04Phase			; //
		GOTO	sa_F04EXIT			; Exit function
;------------------------------------------------------------------------------------------------------------------------
sa_F04PHASE3:	BTFSS	bit_LongDelay1			; Face period expired ?
		GOTO	sa_F04EXIT			; // NO ! - exit function
		MOVLW	001H				; // YES ! - phase 1
		MOVWF	reg_F04Phase			; //
		GOTO	sa_F04EXIT			; Exit function
;------------------------------------------------------------------------------------------------------------------------
sa_F04PHASE4:	BTFSC	bit_LongDelay1			; Face period expired ?
		GOTO	sa_F04PH410			; //  !
		BTFSC	port_MIC			; Sound detected ?
		GOTO	sa_F04PH410			; // NO !
		GOTO	sa_F04EXIT			; Exit function
sa_F04PH410:	MOVLW	001H				; // YES ! - phase 1
		MOVWF	reg_F04Phase			; //
		GOTO	sa_F04EXIT			; Exit function
;------------------------------------------------------------------------------------------------------------------------
sa_F04EXIT						; FUNCTION 04 delimiter
;////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
;
;////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
; FUNCTION 05: Sequence changing by key or microphone & game
		MOVLW	high(sa_F05PHASE)		; Function phase switch
		MOVWF	PCLATH				; //
		MOVF	reg_F05Phase,W			; //
		ADDWF	PCL,F				; //
sa_F05PHASE:	GOTO	sa_F05PHASE0			; //
		GOTO	sa_F05PHASE1			; //
sa_F05PHEND:	GOTO	sa_F05PHASE2			; //
;****** MPASM compiler will issue an error message because of PCLATH overrun !!! ****************************************
	IF	high(sa_F05PHASE) - high(sa_F05PHEND) != 0
		ERROR	"The 256B-boundary of program memory exceeding not allowed !!!"
	ENDIF
;************************************************************************************************************************
sa_F05PHASE0:	BTFSS	bit_Key				; Key pressed ?
		GOTO	sa_F05EXIT			; // NO ! - exit function
		BCF	bit_GameActive			; // YES ! - game inactive
		INCF	reg_SeqCode,F			; Change to the next sequence
		CLRF	reg_F04Phase			; Restart function 4
		MOVLW	007H				; Sequence code >= 0x07 ?
		SUBWF	reg_SeqCode,W			; //
		BTFSC	STATUS,C			; //
		CLRF	reg_SeqCode			; // YES ! - sequence code = 0x00
		INCF	reg_F05Phase,F			; // NO ! - phase 1
		GOTO	sa_F05EXIT			; // YES ! - exit function
;------------------------------------------------------------------------------------------------------------------------
sa_F05PHASE1:	BTFSS	bit_Key				; Key pressed ?
		CLRF	reg_F05Phase			; // NO ! - phase 0
		BTFSC	bit_Hold			; Key hold ?
		INCF	reg_F05Phase,F			; // YES ! - phase 2
		GOTO	sa_F05EXIT			; Exit function
;------------------------------------------------------------------------------------------------------------------------
sa_F05PHASE2:	BTFSC	bit_Hold			; Key released ?
		GOTO	sa_F05EXIT			; // NO ! - exit function
;
		BSF	bit_GameActive			; // YES ! - Set game active
		MOVLW	006H				; Sequence code 0x06
		MOVWF	reg_SeqCode			; //
		MOVLW	con_BasicSpeed			; Initialize game speed
		MOVWF	reg_GameSpeed			; //
		CLRF	reg_OKCounter			; Clear OK counter
		CLRF	reg_F04Phase			; Restart function 4
		CLRF	reg_F05Phase			; Phase 0
		GOTO	sa_F05EXIT			; Exit function
;------------------------------------------------------------------------------------------------------------------------
sa_F05EXIT						; FUNCTION 05 delimiter
;////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
;
;////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
; FUNCTION 06: pseudo-random generator
		MOVLW	014H				; Generate pseudo-random value
		ADDWF	reg_RND1,F			; //
		MOVLW	082H				; //
		ADDWF	reg_RND2,F			; //
		MOVLW	03CH				; //
		ADDWF	reg_RND3,F			; //
		MOVLW	001H				; //
		ADDWF	reg_RND4,F			; //
		MOVF	reg_RND1,W			; //
		XORWF	reg_RND2,W			; //
		XORWF	reg_RND3,W			; //
		XORWF	reg_RND4,W			; //
		ANDLW	0F0H				; //
		MOVWF	reg_Random			; Store pseudo-random value
;------------------------------------------------------------------------------------------------------------------------
sa_F06EXIT						; FUNCTION 06 delimiter
;////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
;
		GOTO	sa_MAINLOOP			; Back to mainloop begin
;
;************************************************************************************************************************
;*		R O M     T A B L E S											*
;************************************************************************************************************************
;
		ORG	0200H
;
; Face decoder is decoding reg_FaceCode --> reg_DriverRow1 & reg_DriverRow2
sa_FACEDECODE:	MOVLW	high(sa_TABLE1BEG)		; Face decoder
		MOVWF	PCLATH				; // conversion of reg_FaceCode --> W
		MOVF	reg_Temp1,W			; //
		ADDWF	PCL,F				; //
;****** MPASM compiler will issue an error message because of PCLATH overrun !!! ****************************************
	IF	high(sa_TABLE1BEG) - high(sa_TABLE1END) != 0
		ERROR	"The 256B-boundary of program memory exceeding not allowed !!!"
	ENDIF
;************************************************************************************************************************
sa_TABLE1BEG:	RETLW	07DH				; Face code 0x00
		RETLW	03DH				; //
		RETLW	067H				; Face code 0x01
		RETLW	027H				; //
		RETLW	06DH				; Face code 0x02
		RETLW	02DH				; //
		RETLW	09DH				; Face code 0x03
		RETLW	0DDH				; //
		RETLW	087H				; Face code 0x04
		RETLW	0C7H				; //
		RETLW	095H				; Face code 0x05
		RETLW	0D5H				; //
		RETLW	0DDH				; Face code 0x06
		RETLW	05DH				; //
		RETLW	0C7H				; Face code 0x07
		RETLW	047H				; //
		RETLW	0DCH				; Face code 0x08
		RETLW	05CH				; //
		RETLW	0D9H				; Face code 0x09
		RETLW	059H				; //
		RETLW	047H				; Face code 0x0A
		RETLW	087H				; //
		RETLW	0D5H				; Face code 0x0B
		RETLW	055H				; //
		RETLW	0CDH				; Face code 0x0C
		RETLW	04DH				; //
		RETLW	047H				; Face code 0x0D
		RETLW	007H				; //
		RETLW	07DH				; Face code 0x0E
		RETLW	027H				; //
		RETLW	000H				; Face code 0x0F
sa_TABLE1END:	RETLW	000H				; //
;
;////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
;
sa_SEQDECODE:	MOVLW	high(sa_TABLE2BEG)		; Sequence decoder
		MOVWF	PCLATH				; // conversion of reg_SeqCode --> W
		MOVF	reg_SeqCode,W			; //
		ADDWF	PCL,F				; //
;****** MPASM compiler will issue an error message because of PCLATH overrun !!! ****************************************
	IF	high(sa_TABLE2BEG) - high(sa_TABLE2END) != 0
		ERROR	"The 256B-boundary of program memory exceeding not allowed !!!"
	ENDIF
;************************************************************************************************************************
sa_TABLE2BEG:	RETLW	low(sa_SEQ00-sa_SEQ00)		; Sequence code 0x00
		RETLW	low(sa_SEQ01-sa_SEQ00)		; Sequence code 0x01
		RETLW	low(sa_SEQ02-sa_SEQ00)		; Sequence code 0x02
		RETLW	low(sa_SEQ03-sa_SEQ00)		; Sequence code 0x03
		RETLW	low(sa_SEQ04-sa_SEQ00)		; Sequence code 0x04
		RETLW	low(sa_SEQ05-sa_SEQ00)		; Sequence code 0x05
sa_TABLE2END:	RETLW	low(sa_SEQ06-sa_SEQ00)		; Sequence code 0x06
;
;////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
;
; low nibble ~ face code, high nibble ~ face period, special code 0x00 ~ EOS (End Of Sequence)
sa_SEQUENCE:	MOVLW	high(sa_TABLE3BEG)		; Sequence decoder
		MOVWF	PCLATH				; // conversion of reg_SeqSymbPtr --> W
		MOVF	reg_SeqSymbPtr,W		; //
		ADDWF	PCL,F				; //
;****** MPASM compiler will issue an error message because of PCLATH overrun !!! ****************************************
	IF	high(sa_TABLE3BEG) - high(sa_TABLE3END) != 0
		ERROR	"The 256B-boundary of program memory exceeding not allowed !!!"
	ENDIF
;************************************************************************************************************************
sa_TABLE3BEG:	
sa_SEQ00:	RETLW	0F7H				; Sequence code 0x00 (sleeping)
		RETLW	04AH				; //
		RETLW	047H				; //
		RETLW	04DH				; //
		RETLW	000H				; // EOS
;
sa_SEQ01:	RETLW	026H				; Sequence code 0x01 (waking-up)
		RETLW	027H				; //
		RETLW	046H				; //
		RETLW	017H				; //
		RETLW	046H				; //
		RETLW	017H				; //
		RETLW	046H				; //
		RETLW	070H				; //
		RETLW	000H				; // EOS
;
sa_SEQ02:	RETLW	011H				; Sequence code 0x02
		RETLW	040H				; //
		RETLW	011H				; //
		RETLW	040H				; //
		RETLW	011H				; //
		RETLW	040H				; //
		RETLW	011H				; //
		RETLW	050H				; //
		RETLW	04EH				; //
		RETLW	050H				; //
		RETLW	000H				; // EOS
;
sa_SEQ03:	RETLW	046H				; Sequence code 0x03
		RETLW	040H				; //
		RETLW	011H				; //
		RETLW	040H				; //
		RETLW	011H				; //
		RETLW	040H				; //
		RETLW	011H				; //
		RETLW	020H				; //
		RETLW	000H				; EOS
;
sa_SEQ04:	RETLW	040H				; Sequence code 0x04
		RETLW	028H				; //
		RETLW	029H				; //
		RETLW	028H				; //
		RETLW	029H				; //
		RETLW	028H				; //
		RETLW	029H				; //
		RETLW	000H				; EOS
;
sa_SEQ05:	RETLW	043H				; Sequence code 0x05
		RETLW	046H				; //
		RETLW	040H				; //
		RETLW	046H				; //
		RETLW	000H				; EOS
;
sa_SEQ06:	RETLW	028H				; Sequence code 0x06
		RETLW	029H				; //
		RETLW	02BH				; //
		RETLW	02CH				; //
sa_TABLE3END:	RETLW	000H				; EOS
;
;////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
;
;************************************************************************************************************************
;*		F A C E     G R A P H I C S										*
;************************************************************************************************************************
;
; LED assignment:
;
;					bit	| 7.| 6.| 5.| 4.| 3.| 2.| 1.| 0.|
;	  D	   I		-------------------------------------------------
;	A B C	 F G H		reg_DriverRow1	| P | O | M | J | I | H | G | F |
;	  E	   J		-------------------------------------------------
;				reg_DriverRow2	| L | N | K | E | D | C | B | A | 
;
;
;	  K  L  L  M
;	  N  O  O  P
;
;------------------------------------------------------------------------------------------------------------------------
; Face codes (reg_DriverRow1,reg_DriverRow2):
;
;	0x00 (0x7D,0x3D)		0x01 (0x67,0x27)		0x02 (0x6D,0x2D)		0x03 (0x9D,0xDD)
;
;	  X	   X			  . 	   . 			  X	   X			  X	   X
;	X .  X	 X . X			X X X	 X X X			X . X	 X . X			X . X	 X . X
;	  X	   X			  . 	   . 			   	    			  X	   X
;
;
;
;	  X  .  .  X			  X  .  .  X			  X  .  .  X			  .  X  X  .
;	  .  X  X  . 			  .  X  X .  			  .  X  X  .			  X  .  .  X
;
;------------------------------------------------------------------------------------------------------------------------
;
;	0x04 (0x87,0xC7)		0x05 (0x95,0xD5)		0x06 (0xDD,0x5D)		0x07 (0xC7,0x47)
;
;	  .	   .			  .	   .			  X	   X			  .	   .
;	X X X	 X X X			X . X	 X . X			X . X	 X . X			X X X	 X X X
;	  .	   .			  X	   X			  X	   X			  .	   .
;
;
;
;	  .  X  X  .			  .  X  X  .			  .  .  .  .			  .  .  .  .
;	  X  .  .  X			  X  .  .  X			  X  X  X  X			  X  X  X  X
;
;------------------------------------------------------------------------------------------------------------------------
;
;	0x08 (0xDE,0x5E)		0x09 (0xDB,0x5B)		0x0A (0x47,0x87)		0x0B (0xD5,0x55)
;
;	  X	   X			  X	   X			  .	   .			  .	   .
;	. . X	 . . X			X . .	 X . .			X X X	 X X X			X . X	 X . X
;	  X	   X			  X	   X			  .	   .			  X	   X
;
;
;
;	  .  .  .  .			  .  .  .  .			  .  X  X  .			  .  .  .  .
;	  X  X  X  X			  X  X  X  X			  .  X  X  .			  X  X  X  X
;
;------------------------------------------------------------------------------------------------------------------------
;
;	0x0C (0xCD,0x4D)		0x0D (0x47,0x07)		0x0E (0x67,0x3D)		0x0F (0x00,0x00)
;
;	  X	   X			  .	   .			  X	   .			  .	   .
;	X . X	 X . X			X X X	 X X X			X . X	 X X X			. . .	 . . .
;	  .	   .			  .	   .			  X	   .			  .	   .
;
;
;
;	  .  .  .  .			  .  .  .  .			  X  .  .  X			  .  .  .  .
;	  X  X  X  X			  .  X  X  .			  .  X  X  .			  .  .  .  .
;
;
;************************************************************************************************************************
;*		A P P L I C A T I O N     E N D										*
;************************************************************************************************************************
;
;
;
		END

